-- compute
#version 460

layout( local_size_x = 64 ) in;

uniform mat4	  mv;
uniform mat4	  proj;
uniform float     width;
uniform float     height;
uniform vec3      extent;
uniform float     zNear;
uniform float     zFar;
uniform sampler2D hiZBuffer;

struct DrawElementsIndirectCommand
{
    uint    count;
    uint    instanceCount;
    uint    firstIndex;
    uint    baseVertex;
    uint    baseInstance;
};
 
layout(std430, binding = 0) buffer Pos 
{
	vec4 position [];
};

layout(std430, binding = 1) buffer Cmd
{
	DrawElementsIndirectCommand cmds [];
};

vec4 boundingBox [8];

bool	frustumCull ( in vec3 instPos, in vec3 ext )
{
    mat4 mvp = proj * mv;

    	// create the bounding box of the object
    boundingBox [0] = mvp * vec4 ( instPos + vec3 (  ext.x,  ext.y,  ext.z ), 1.0 );
    boundingBox [1] = mvp * vec4 ( instPos + vec3 ( -ext.x,  ext.y,  ext.z ), 1.0 );
    boundingBox [2] = mvp * vec4 ( instPos + vec3 (  ext.x, -ext.y,  ext.z ), 1.0 );
    boundingBox [3] = mvp * vec4 ( instPos + vec3 ( -ext.x, -ext.y,  ext.z ), 1.0 );
    boundingBox [4] = mvp * vec4 ( instPos + vec3 (  ext.x,  ext.y, -ext.z ), 1.0 );
    boundingBox [5] = mvp * vec4 ( instPos + vec3 ( -ext.x,  ext.y, -ext.z ), 1.0 );
    boundingBox [6] = mvp * vec4 ( instPos + vec3 (  ext.x, -ext.y, -ext.z ), 1.0 );
    boundingBox [7] = mvp * vec4 ( instPos + vec3 ( -ext.x, -ext.y, -ext.z ), 1.0 );
   
    	// check how the bounding box resides regarding to the view frustum
    int outOfBound [6] = int[6]( 0, 0, 0, 0, 0, 0 );

	// check very box vertex against each plane
    for ( int i = 0; i < 8; i++ )
    {
        if ( boundingBox[i].x >  boundingBox[i].w ) 
            outOfBound [0]++;

        if ( boundingBox [i].x < -boundingBox [i].w ) 
            outOfBound [1]++;

        if ( boundingBox [i].y >  boundingBox [i].w ) 
            outOfBound [2]++;

        if ( boundingBox [i].y < -boundingBox [i].w ) 
            outOfBound [3]++;

        if ( boundingBox [i].z >  boundingBox [i].w ) 
            outOfBound [4]++;

        if ( boundingBox [i].z < -boundingBox [i].w ) 
            outOfBound [5]++;
    }

    for ( int i = 0; i < 6; i++ )
        if ( outOfBound [i] == 8 ) 
            return false;

    return true;
}

float   linearDepth ( float d )
{
  return (2.0 * zNear) / (zFar + zNear - d * (zFar - zNear));
}

bool	hiZHidden ( in vec3 instPos, in vec3 ext )
{
		// first do frustum cull
	if ( !frustumCull ( instPos, ext ) ) 
		return true;
	
		// perform perspective division for the bounding box
		// npw boundingBox [i] in clip space [-1,1]^3
	for ( int i = 0; i < 8; i++ )
		boundingBox [i].xyz /= boundingBox [i].w;
	
	vec2 boundingRect [2];

		// calculate screen space bounding rectangle
		// convert xy from clip space to [0,1]^2
	boundingRect[0].x = min ( min ( min ( boundingBox [0].x, boundingBox [1].x ),
								    min ( boundingBox [2].x, boundingBox [3].x ) ),
							  min ( min ( boundingBox [4].x, boundingBox [5].x ),
								    min ( boundingBox [6].x, boundingBox [7].x ) ) ) / 2.0 + 0.5;
                                    
	boundingRect[0].y = min ( min ( min ( boundingBox [0].y, boundingBox [1].y ),
								    min ( boundingBox [2].y, boundingBox [3].y ) ),
							  min ( min ( boundingBox [4].y, boundingBox [5].y ),
								    min ( boundingBox [6].y, boundingBox [7].y ) ) ) / 2.0 + 0.5;
                                    
	boundingRect[1].x = max ( max ( max ( boundingBox [0].x, boundingBox [1].x ),
								    max ( boundingBox [2].x, boundingBox [3].x ) ),
							  max ( max ( boundingBox [4].x, boundingBox [5].x ),
								    max ( boundingBox [6].x, boundingBox [7].x ) ) ) / 2.0 + 0.5;
                                    
	boundingRect[1].y = max ( max ( max ( boundingBox [0].y, boundingBox [1].y ),
								    max ( boundingBox [2].y, boundingBox [3].y ) ),
							  max ( max ( boundingBox [4].y, boundingBox [5].y ),
								    max ( boundingBox [6].y, boundingBox [7].y ) ) ) / 2.0 + 0.5;

		// then depth value of the front-most (nearest) point
	float instanceDepth = min ( min ( min ( boundingBox [0].z, boundingBox  [1].z ),
									  min ( boundingBox [2].z,  boundingBox [3].z ) ),
							    min ( min ( boundingBox [4].z,  boundingBox [5].z ),
								 	  min ( boundingBox [6].z,  boundingBox [7].z ) ) );

		// convert min depth from [-1,1] to [0,1]
	instanceDepth = instanceDepth / 2.0 + 0.5;

        // calculate the bounding rectangle size in viewport coordinates
	float viewSizeX = (boundingRect [1].x - boundingRect [0].x) * width;
	float viewSizeY = (boundingRect [1].y - boundingRect [0].y) * height;
	
        // calculate texture LOD used for lookup in the depth buffer texture
	float LOD = ceil ( log2 ( max ( viewSizeX, viewSizeY ) / 2.0 ) );
	
        // finally fetch the depth texture using explicit LOD lookups
	vec4 samples;
    
	samples.x = textureLod ( hiZBuffer, vec2 ( boundingRect [0].x, boundingRect [0].y ), LOD ).x;
	samples.y = textureLod ( hiZBuffer, vec2 ( boundingRect [0].x, boundingRect [1].y ), LOD ).x;
	samples.z = textureLod ( hiZBuffer, vec2 ( boundingRect [1].x, boundingRect [1].y ), LOD ).x;
	samples.w = textureLod ( hiZBuffer, vec2 ( boundingRect [1].x, boundingRect [0].y ), LOD ).x;
    
		// get max value depth from buffer
	float maxDepth = max ( max ( samples.x, samples.y ), max ( samples.z, samples.w ) );
	
		// object is hidden if it is farther than buffer max depth
	bool	isHidden = instanceDepth >= maxDepth;

	return isHidden;
}

void main() 
{
	uint idx = gl_GlobalInvocationID.x;
    
	if ( idx < position.length () )
	{
		vec3 p   = position [idx].xyz;
    
		cmds [idx].instanceCount = hiZHidden ( p, extent ) ? 0 : 1;
	}
}
