//
// Hierarchical z-buffer visibility example 
//
// Author: Alexey V. Boreskov <steps3d@gmail.com>, <steps3d@narod.ru>
//

#include	<algorithm>
#include	"GlutCameraWindow.h"
#include	"Program.h"
#include	"BasicMesh.h"
#include	"Texture.h"
#include	"Framebuffer.h"
#include	"ScreenQuad.h"
#include	"randUtils.h"

#define	SIZE	1000.0f

struct DrawElementsIndirectCommand 
{ 
    GLuint  count;              // number of elements to be rendered
    GLuint  instanceCount;      // number of instances of indexed geometry to be rendered
    GLuint  firstIndex; 
    GLint   baseVertex; 
    GLuint  baseInstance; 
};

class	HiZDemoWindow : public GlutCameraWindow
{
	Program		    program, program2, downsampleProgram, cullProgram, resetProgram, mapProgram;
	FrameBuffer     fb;
	ScreenQuad	    quad;
	VertexBuffer    posBuf;             // vector of vec4
	VertexBuffer    drawBuf;            // vector of DrawElementsIndirectCommand
	int             numLevels = 1;      // # if levels in mipmap pyramid
	int             numModels = 100000;
	Texture       * depthMap  = nullptr;
	Texture       * colorMap  = nullptr;
	BasicMesh     * room      = nullptr;
	BasicMesh     *	knot      = nullptr;
	bool            cull      = false;
    
	Texture 	decalMap;
	Texture 	stoneMap;

	std::vector<BasicMesh *>	occluders;
    
public:
	HiZDemoWindow () : GlutCameraWindow ( 100, 100, 900, 900, "Hierarchical z-buffer visibility" ) 
	{
		createBuffer ( fb );		// create framebuffer for pyramid building

		if ( !program.loadProgram ( "decal.glsl" ) )
			exit ( "Error loading shader: %s\n", program.getLog ().c_str () );
		
		program.bind       ();
		program.setTexture ( "imageMap", 0 );
		program.unbind     ();

		if ( !program2.loadProgram ( "decal-pos.glsl" ) )
			exit ( "Error loading shader: %s\n", program2.getLog ().c_str () );
		
		program2.bind       ();
		program2.setTexture ( "imageMap", 0 );
		program2.unbind     ();
		
		if ( !downsampleProgram.loadProgram ( "downsample.glsl" ) )
			exit ( "Error loading shader: %s\n", downsampleProgram.getLog ().c_str () );
		
		downsampleProgram.bind       ();
		downsampleProgram.setTexture ( "lastMip", 0 );
		downsampleProgram.unbind     ();
		
		if ( !cullProgram.loadProgram ( "cull.glsl" ) )
			exit ( "Error loading shader: %s\n", cullProgram.getLog ().c_str () );
		
		cullProgram.bind       ();
		cullProgram.setTexture ( "hiZBuffer", 0 );
		cullProgram.unbind     ();
		
		if ( !resetProgram.loadProgram ( "reset-cull.glsl" ) )
			exit ( "Error loading shader: %s\n", resetProgram.getLog ().c_str () );

		if ( !mapProgram.loadProgram ( "depth.glsl" ) )
			exit ( "Error loading shader: %s\n", mapProgram.getLog ().c_str () );
		
		mapProgram.bind       ();
		mapProgram.setTexture ( "depthMap", 0 );
		mapProgram.unbind     ();

		createBoxes ();

		decalMap.load2D ( "../../Textures/oak.jpg" );
		stoneMap.load2D ( "../../Textures/block.jpg" );

		camera.moveTo ( glm::vec3 ( -0.5, 0.5, 5 ) );

            // we need to alloc memory for mipmap pyramid or 
            // will not be able to render to them
        depthMap->bind         ();
        depthMap->buildMipmaps ();
        depthMap->setWrapAll   ( GL_CLAMP_TO_EDGE );
        depthMap->setFilter    ( GL_NEAREST/*_MIPMAP_NEAREST*/, GL_NEAREST );
        depthMap->unbind       ();

        setCaption ( std::string ( "Hierarchical z-buffer " ) + ( cull ? "on" : "off" ) );
	}

	void redisplay () override
	{
        renderOccluders   ();
        buildDepthPyramid ();
        buildVisibleList  ();
        renderVisibles    ();

		drawMap ( 10,  10, 0 );
		drawMap ( 160, 10, 2 );
		drawMap ( 310, 10, 4 );
		drawMap ( 460, 10, 6 );
		drawMap ( 610, 10, 7 );
		drawMap ( 760, 10, 8 );
	}

	void	keyTyped ( unsigned char key, int modifiers, int x, int y ) override
	{
		GlutWindow::keyTyped ( key, modifiers, x, y );

		if ( key == 27 || key == 'q' || key == 'Q' )	// quit requested
			::exit ( 0 );
			
		glm::vec3	mv ( 0 );

		if ( key == 'w' || key == 'W' )
			mv = step * camera.getViewDir ();
		else
		if ( key == 'x' || key == 'X' )
			mv = -sideStep * camera.getViewDir ();
		else
		if ( key == 'a' || key == 'A' )
			mv = -step * camera.getSideDir ();
		else
		if ( key == 'd' || key == 'D' )
			mv = sideStep * camera.getSideDir ();
        else
        if ( key == ' ' )
        {
            cull = !cull;
            
			resetCullingMode ();
        }

		camera.moveBy ( glm::vec3 ( mv.x, 0.0f, mv.z ) );
	}

	void mouseWheel ( int wheel, int dir, int x, int y ) override {}

protected:
	void    createBoxes ()
	{
		room = createBox  ( glm::vec3 ( -SIZE, -0.1, -SIZE ),  glm::vec3 ( 2*SIZE, 3, 2*SIZE ) );  // stoneMap, false
		//knot = createKnot ( glm::vec3 ( 0, 1.5, 0 ), 0.4, 0.2, 30, 30 );
		knot = createSphere  ( glm::vec3 ( 0, 0, 0 ), 0.3, 30, 30 );
        
		for ( int i = -SIZE / 4; i <= SIZE / 4; i += 4 )
			for ( int j = -SIZE / 4; j <= SIZE / 4; j += 4 )
			{
				glm::vec3	pos  ( i + 2*randUniform ( -1, 1 ), -0.1f, j + 2*randUniform ( -1, 1 ) );
				glm::vec3	size ( 1 + 2*randUniform (), 3, 1 + 2 * randUniform () );

				occluders.push_back ( createBox ( pos, size ) );
			}

		std::vector<glm::vec4>	pos;

		for ( int i = 0; i < numModels; i++ )
		{
			glm::vec4	p ( 0.2f*randUniform (-SIZE, SIZE), 0.2f*randUniform (1,3), 0.2f*randUniform (-SIZE, SIZE), 1 );

			pos.push_back ( p );
		}

		posBuf.create    ();
		posBuf.bindBase  ( GL_SHADER_STORAGE_BUFFER, 0 );
		posBuf.setData   ( pos.size () * sizeof ( glm::vec4 ), pos.data (), GL_STATIC_DRAW );
		posBuf.unbind    ();
        
        DrawElementsIndirectCommand cmd;
        
        cmd.count         = 3*knot->getNumTriangles ();
        cmd.instanceCount = 1;
        cmd.firstIndex    = 0;
        cmd.baseVertex    = 0;
        cmd.baseInstance  = 0;
        
        std::vector<DrawElementsIndirectCommand>    cmds ( numModels );
        
        for ( int i = 0; i < numModels; i++ )
            cmds [i] = cmd;
        
        drawBuf.create   ();
        drawBuf.bindBase ( GL_SHADER_STORAGE_BUFFER, 1 );
        drawBuf.setData  ( numModels * sizeof ( DrawElementsIndirectCommand ), cmds.data (), GL_STATIC_DRAW );
        drawBuf.unbind   ();
	}
    
	void	createBuffer ( FrameBuffer& fb )
	{
		fb.create             ( getWidth (), getHeight (), FrameBuffer::depth24 );
		fb.bind               ();
		fb.attachColorTexture ( colorMap = fb.createColorTexture () );
		fb.attachDepthTexture ( depthMap = fb.createDepthTexture () );
		
		if ( !fb.isOk () )
			exit ( "Error with fb\n" );
			
		fb.unbind ();	
	}

    void    renderOccluders ()
    {
		glEnable ( GL_DEPTH_TEST  );
		glClear  ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		
		program.bind             ();
		program.setUniformMatrix ( "proj",  camera.getProjection () );
		program.setUniformMatrix ( "mv",    camera.getModelView  () );
		
		decalMap.bind ();
		room -> render ();
		decalMap.unbind ();

		stoneMap.bind ();

		for ( auto * p : occluders )
			p -> render ();

		stoneMap.unbind ();
		
		program.unbind ();
    }
    
    void    buildDepthPyramid ()
    {
        downsampleProgram.bind ();
        
		depthMap->bind    ();

            // copy depth buffer contents
		glCopyTexImage2D ( depthMap->getTarget (), 0, GL_DEPTH_COMPONENT, 0, 0, depthMap->getWidth (), depthMap->getHeight (), 0 );

			// prepare rendering to framebuffer
        fb.bind ();

            // build mipmap pyramid for 
			// start post processing, we don't need depth test
    	glColorMask ( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE );
        glDepthFunc ( GL_ALWAYS );        

            // calculate the number of mipmap levels for NPOT texture
        int w         = getWidth  ();
        int h         = getHeight ();
        int numLevels = 1 + (int)floorf ( log2f ( fmaxf ( w, h ) ) );

        for ( int i = 1; i < numLevels; i++ )
        {
            downsampleProgram.setUniformInt ( "lastMipWidth",  w );
            downsampleProgram.setUniformInt ( "lastMipHeight", h );
            
                // calculate next viewport size
			w = std::max ( 1, w / 2 );
			h = std::max ( 1, h / 2 );
            
                // set render area
            glViewport ( 0, 0, w, h );
            
                // bind next level for rendering but first restrict fetches only to previous level
            glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, i - 1 );
            glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL,  i - 1 );

            fb.attachDepthTexture ( depthMap, i );

			// draw full-screen quad
            quad.render ();
        }
        
        downsampleProgram.unbind ();
        
            // reset mipmap level range for the depth image
        glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0 );
        glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL,  numLevels - 1 );
        
        glColorMask ( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE );
		glDepthFunc ( GL_LEQUAL );

            // restore viewport
        glViewport ( 0, 0, getWidth (), getHeight () );

           // restore depth mip 0 as framebuffer binding
        fb.attachDepthTexture ( depthMap, 0 );
        fb.unbind ();
    }

    void    buildVisibleList ()
    {
        auto   sz = knot -> getBox ().getSize () * 0.5f;

        if ( cull )
        {
            depthMap -> bind      ( 0 );
			depthMap -> setFilter ( GL_NEAREST_MIPMAP_NEAREST, GL_NEAREST );

            cullProgram.bind ();
            cullProgram.setUniformMatrix ( "proj",   camera.getProjection () );
            cullProgram.setUniformMatrix ( "mv",     camera.getModelView  () );
            cullProgram.setUniformFloat  ( "width",  getWidth  () );
            cullProgram.setUniformFloat  ( "height", getHeight () );
            cullProgram.setUniformVector ( "extent", sz );
            cullProgram.setUniformFloat  ( "zNear",  getCamera ().getZNear () );
            cullProgram.setUniformFloat  ( "zFar",   getCamera ().getZFar  () );

            drawBuf.bindBase ( GL_SHADER_STORAGE_BUFFER, 1 );
            posBuf.bindBase  ( GL_SHADER_STORAGE_BUFFER, 0 );
            
                // run one thread of cull shader per each model
            glDispatchCompute ( (numModels + 63) / 64, 1, 1 );
            
			depthMap -> setFilter    ( GL_NEAREST, GL_NEAREST );
            cullProgram.unbind ();
        }      
    }
        
    void    renderVisibles ()
    {       
        program2.bind ();
		program2.setUniformMatrix ( "proj",  camera.getProjection () );
		program2.setUniformMatrix ( "mv",    camera.getModelView  () );

        knot->getVao ().bind ();
        
        drawBuf.bind    ( GL_DRAW_INDIRECT_BUFFER );
		posBuf.bindBase ( GL_SHADER_STORAGE_BUFFER, 0 );
        
        glMultiDrawElementsIndirect ( GL_TRIANGLES, GL_UNSIGNED_INT, 0, numModels, 0 );
        
        knot->getVao ().unbind ();
        
        program2.unbind ();
    }

	void drawMap ( int x, int y, int level )
	{
		glViewport ( x, y, 128, 128 );

		mapProgram.bind            ();
		mapProgram.setUniformFloat ( "zNear",  getCamera ().getZNear () );
		mapProgram.setUniformFloat ( "zFar",   getCamera ().getZFar  () );
		mapProgram.setUniformInt   ( "lod",    level );
		depthMap->bind ();
		depthMap->setFilter    ( GL_NEAREST_MIPMAP_NEAREST, GL_NEAREST );
		quad.render();
		depthMap->setFilter    ( GL_NEAREST, GL_NEAREST );
		mapProgram.unbind ();

		glViewport ( 0, 0, getWidth (), getHeight () );
	}

		// reset visibility buffer to all visible
	void resetCullingMode ()
	{
		setCaption ( std::string ( "Hierarchical z-buffer " ) + ( cull ? "on" : "off" ) );

		resetProgram.bind ();
		drawBuf.bindBase  ( GL_SHADER_STORAGE_BUFFER, 1 );

		glDispatchCompute ( (numModels + 63) / 64, 1, 1 );

		resetProgram.unbind ();
	}
};

int main ( int argc, char * argv [] )
{
	GlutWindow::init ( argc, argv, 4, 6 );
	
	HiZDemoWindow	win;
	
	return GlutWindow::run ();
}
